/* Workaround for PowerPC 64-bit _Complex float passing ABI change
   Copyright (C) 2005 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Jakub Jelinek <jakub@redhat.com>, 2005.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

/* Build with:
   sed -n '/^GLIBC_2.3/,/^};$/p' libComplexFloat-GCC_3.2.c \
     > libComplexFloat-GCC_3.2.map
   gcc -m64 -O2 -fno-builtin -fpic -shared -o libComplexFloat-GCC_3.2.so \
     libComplexFloat-GCC_3.2.c \
     -Wl,--version-script,libComplexFloat-GCC_3.2.map -ldl -lm
   Test with:
   gcc -m64 -O0 -fno-builtin -DTEST -o libComplexFloat-GCC_3.2{,.c} -lm
   if ! ./libComplexFloat-GCC_3.2; then echo gcc test failed; exit 1; fi
   gcc32 -m64 -O0 -fno-builtin -DTEST -o libComplexFloat-GCC_3.2{,.c} -lm
   if ! LD_PRELOAD=./libComplexFloat-GCC_3.2.so ./libComplexFloat-GCC_3.2; \
     then echo gcc32 test failed; exit 1; fi
 
GLIBC_2.3 {
  global:
    cabsf; cargf; cimagf; crealf; cacosf; casinf; catanf; ccosf; csinf;
    ctanf; cacoshf; casinhf; catanhf; ccoshf; csinhf; ctanhf; cexpf;
    clogf; clog10f; __clog10f; csqrtf; conjf; cprojf; cpowf;
  local:
    *;
};
*/

#ifndef TEST

#define _GNU_SOURCE 1
#include <dlfcn.h>

typedef float (*vc) (_Complex float); 
typedef _Complex float (*cc) (_Complex float); 
typedef _Complex float (*ccc) (_Complex float, _Complex float); 
typedef union { unsigned long l; float cf[2]; } U;

static vc fcabs, fcarg, fcimag, fcreal;
static cc fcacos, fcasin, fcatan, fccos, fcsin, fctan;
static cc fcacosh, fcasinh, fcatanh, fccosh, fcsinh, fctanh;
static cc fcexp, fclog, fclog10, f__clog10, fcsqrt, fconj, fcproj;
static ccc fcpow;

static void __attribute__((constructor))
_init (void)
{
#define L(name) f##name = dlvsym (RTLD_NEXT, #name "f", "GLIBC_2.3");
  L(cabs)
  L(carg)
  L(cimag)
  L(creal)
  L(cacos)
  L(casin)
  L(catan)
  L(ccos)
  L(csin)
  L(ctan)
  L(cacosh)
  L(casinh)
  L(catanh)
  L(ccosh)
  L(csinh)
  L(ctanh)
  L(cexp)
  L(clog)
  L(clog10)
  L(__clog10)
  L(csqrt)
  L(conj)
  L(cproj)
  L(cpow)
}

#define VC(name) \
float							\
__attribute__((weak))					\
name##f (unsigned long x)				\
{							\
  U u;							\
  u.l = x;						\
  return f##name (u.cf[0] + u.cf[1] * 1.0iF);		\
}

#define CC(name) \
unsigned long						\
__attribute__((weak))					\
name##f (unsigned long x)				\
{							\
  U u;							\
  u.l = x;						\
  _Complex float ret;					\
  ret = f##name (u.cf[0] + u.cf[1] * 1.0iF);		\
  u.cf[0] = __real__ ret;				\
  u.cf[1] = __imag__ ret;				\
  return u.l;						\
}

#define CCC(name) \
unsigned long						\
__attribute__((weak))					\
name##f (unsigned long x, unsigned long y)		\
{							\
  U u, v;						\
  u.l = x;						\
  v.l = y;						\
  _Complex float ret;					\
  ret = f##name (u.cf[0] + u.cf[1] * 1.0iF,		\
		 v.cf[0] + v.cf[1] * 1.0iF);		\
  u.cf[0] = __real__ ret;				\
  u.cf[1] = __imag__ ret;				\
  return u.l;						\
}

VC(cabs)
VC(carg)
VC(cimag)
VC(creal)
CC(cacos)
CC(casin)
CC(catan)
CC(ccos)
CC(csin)
CC(ctan)
CC(cacosh)
CC(casinh)
CC(catanh)
CC(ccosh)
CC(csinh)
CC(ctanh)
CC(cexp)
CC(clog)
CC(clog10)
CC(__clog10)
CC(csqrt)
CC(conj)
CC(cproj)
CCC(cpow)

#else

#undef TEST
#define _GNU_SOURCE 1
#include <complex.h>
#include <stdlib.h>

int main (void)
{
#define TEST(x, expected) \
  do								\
    {								\
      if (x < expected - 0.0001f || x > expected + 0.0001f)	\
	abort ();						\
    }								\
  while (0);
#define CTEST(x, real, imag) \
  TEST (__real__ x, real)					\
  TEST (__imag__ x, imag)

  _Complex float f, ret;
  float g;
  __real__ f = 1.0;
  __imag__ f = 2.0;
  g = cabsf (f);
TEST (g, 2.236068f)
  g = cargf (f);
TEST (g, 1.107149f)
  g = cimagf (f);
TEST (g, 2.000000f)
  g = crealf (f);
TEST (g, 1.000000f)
  ret = cacosf (f);
CTEST (ret, 1.143718f, -1.528570f)
  ret = casinf (f);
CTEST (ret, 0.427078f, 1.528570f)
  ret = catanf (f);
CTEST (ret, 1.338972f, 0.402359f)
  ret = ccosf (f);
CTEST (ret, 2.032723f, -3.051898f)
  ret = csinf (f);
CTEST (ret, 3.165779f, 1.959601f)
  ret = ctanf (f);
CTEST (ret, 0.033813f, 1.014794f)
  ret = cacoshf (f);
CTEST (ret, 1.528571f, 1.143718f)
  ret = casinhf (f);
CTEST (ret, 1.469352f, 1.063440f)
  ret = catanhf (f);
CTEST (ret, 0.173287f, 1.178097f)
  ret = cexpf (f);
CTEST (ret, -1.131204f, 2.471727f)
  ret = clogf (f);
CTEST (ret, 0.804719f, 1.107149f)
  ret = clog10f (f);
CTEST (ret, 0.349485f, 0.480829f)
  ret = csqrtf (f);
CTEST (ret, 1.272020f, 0.786151f)
  ret = conjf (f);
CTEST (ret, 1.000000f, -2.000000f)
  ret = cprojf (f);
CTEST (ret, 0.333333f, 0.666667f)
  ret = cpowf (f, f);
CTEST (ret, -0.222517f, 0.100709f)
  return 0;
}

#endif
